import * as Konva from "konva";
import { ICircle } from "./circle";
import { IArrow } from "./arrow";
import { IImage } from "./image";
import { IPolygon } from "./polygon";
import { IRectangle } from "./rectangle";
import { keys, isNil } from "lodash";

export interface IThisParam {
  layer: Konva.Layer;
  stage: Konva.Stage;
  scale: number;
}

type IShape<T> = T extends Konva.Arrow
  ? Partial<IArrow>
  : T extends Konva.Circle
    ? Partial<ICircle>
    : T extends Konva.Image
      ? Partial<IImage>
      : T extends Konva.Rect
        ? Partial<IRectangle>
        : T extends Konva.Shape ? Partial<IPolygon> : {};

enum EventTypes {
  "mousedown",
  "mouseup",
  "touchstart",
  "touchend"
}

export class R2dObject {
  private readonly nativeObject: Konva.Shape;
  public getNativeObject = () => this.nativeObject;
  private readonly layer: Konva.Layer;
  private readonly stage: Konva.Stage;
  private readonly scale: number;

  public constructor(nativeObject: Konva.Shape, thisParam: IThisParam) {
    this.nativeObject = nativeObject;
    this.layer = thisParam.layer;
    this.stage = thisParam.stage;
    this.scale = thisParam.scale;
  }

  public destroy() {
    this.nativeObject.destroy();
  }

  public intersects(object: R2dObject) {
    return this.nativeObject.intersects(object.getNativeObject().position());
  }

  public show() {
    this.nativeObject.show();
  }

  public hide() {
    this.nativeObject.hide();
  }

  public on(
    events: Array<keyof typeof EventTypes>,
    callback: (e: { x: number; y: number }) => void
  ) {
    this.getNativeObject().on(events.join(" "), () => {
      callback({
        x:
          (this.stage.height() / 2 -
            (this.stage.getPointerPosition().y -
              this.stage.getAbsolutePosition().y) /
              this.stage.getAbsoluteScale().y) *
          (1 / this.scale),
        y:
          (this.stage.width() / 2 -
            (this.stage.getPointerPosition().x -
              this.stage.getAbsolutePosition().x) /
              this.stage.getAbsoluteScale().x) *
          (1 / this.scale)
      });
    });
  }

  public off(events: Array<keyof typeof EventTypes>) {
    this.getNativeObject().off(events.join(" "));
  }

  protected internalUpdateAppearance<T>(args: IShape<T>) {
    for (const key of keys(args)) {
      const value = args[key as keyof IShape<T>];
      if (isNil(value)) {
        throw new Error(`${key} cannot be null or undefined`);
      }
      switch (key) {
        case "color": {
          (this.getNativeObject() as Konva.Shape).fill(
            (args as IPolygon)[key]!
          );
          break;
        }
        case "border": {
          const border = (args as IPolygon)[key]!;
          (this.getNativeObject() as Konva.Shape).strokeWidth(border);
          (this.getNativeObject() as Konva.Shape).strokeEnabled(border !== 0);
          break;
        }
        case "borderColor": {
          (this.getNativeObject() as Konva.Shape).stroke(
            (args as IPolygon)[key]!
          );
          break;
        }
        case "opacity": {
          (this.getNativeObject() as Konva.Shape).opacity(
            (args as IPolygon)[key]!
          );
          break;
        }
        case "radius": {
          (this.getNativeObject() as Konva.Circle).radius(
            (args as ICircle)[key]
          );
          break;
        }
        case "pointerWidth": {
          (this.getNativeObject() as Konva.Arrow).pointerWidth(
            (args as IArrow)[key]
          );
          break;
        }
        case "pointerLength": {
          (this.getNativeObject() as Konva.Arrow).pointerLength(
            (args as IArrow)[key]
          );
          break;
        }
        case "width": {
          (this.getNativeObject() as Konva.Image | Konva.Rect).width(
            (args as IImage)[key]!
          );
          break;
        }
        case "height": {
          (this.getNativeObject() as Konva.Image | Konva.Rect).height(
            (args as IImage)[key]!
          );
          break;
        }
      }
    }
    this.layer.batchDraw();
  }
}
