import { isNil } from "lodash";
import * as Konva from "konva";
import { IThisParam, R2dObject } from "./R2dObject";

export interface IImage {
  width?: number;
  height?: number;
  border?: number;
  borderColor?: string;
  opacity?: number;
}

export type IImageSync = IImage & { image: HTMLImageElement };
export type IImageAsync = IImage & { imageURL: string };

export interface IImageAsyncReturnSuccess {
  image: ImageObject;
}

export class ImageObject extends R2dObject {
  public readonly identifier = "image";
  constructor(args: Konva.ImageConfig, thisParam: IThisParam) {
    super(new Konva.Image(args), thisParam);
  }
  public updateAppearance(args: Partial<IImage>) {
    super.internalUpdateAppearance<Konva.Image>(args);
  }
}

export function imageSync(this: IThisParam, args: IImageSync) {
  return new ImageObject(
    {
      image: args.image,
      width: !isNil(args.width)
        ? args.width
        : !isNil(args.height)
          ? (args.image.naturalWidth / args.image.naturalHeight) * args.height
          : args.image.naturalWidth,
      height: !isNil(args.height)
        ? args.height
        : !isNil(args.width)
          ? (args.image.naturalHeight / args.image.naturalWidth) * args.width
          : args.image.naturalHeight,
      offsetX:
        (!isNil(args.width)
          ? args.width
          : !isNil(args.height)
            ? (args.image.naturalWidth / args.image.naturalHeight) * args.height
            : args.image.naturalWidth) / 2,
      offsetY:
        (!isNil(args.height)
          ? args.height
          : !isNil(args.width)
            ? (args.image.naturalHeight / args.image.naturalWidth) * args.width
            : args.image.naturalHeight) / 2,
      strokeEnabled: isNil(args.border) ? false : args.border !== 0,
      strokeWidth: !isNil(args.border) ? args.border : 0,
      stroke: !isNil(args.borderColor) ? args.borderColor : "black",
      opacity: !isNil(args.opacity) ? args.opacity : 1,
      x: this.stage.width() / 2,
      y: this.stage.height() / 2
    },
    this
  );
}

export async function imageAsync(
  this: { layer: Konva.Layer },
  args: IImageAsync
) {
  const image = new Image();
  image.src = args.imageURL;
  // Todo: use translated width and height
  return new Promise<IImageAsyncReturnSuccess>(resolve => {
    image.onload = () =>
      resolve({
        image: imageSync.bind(this)({
          image,
          width: args.width,
          height: args.height,
          border: args.border,
          borderColor: args.borderColor,
          opacity: args.opacity
        })
      });
    image.onerror = () => {
      throw new Error("error loading image");
    };
  });
}
