import * as Konva from "konva";
import * as ROSLib from "roslib";
import {
  animateToPoint,
  getContainerDimensions,
  isRosBridgeURL,
  quaternionToEulerAngle,
  ZoomControl
} from "./utils";
import circle, { CircleObject, ICircle } from "./helpers/circle";
import { isNil, map } from "lodash";
import rectangle, { IRectangle, RectangleObject } from "./helpers/rectangle";
import arrow, { ArrowObject, IArrow } from "./helpers/arrow";
import {
  IImageAsync,
  IImageAsyncReturnSuccess,
  IImageSync,
  ImageObject,
  imageAsync,
  imageSync
} from "./helpers/image";
import drawableMap, { IMap, IMapReturnSuccess } from "./helpers/map";
import polygon, { IPolygon, PolygonObject } from "./helpers/polygon";
import { R2dObject } from "./helpers/R2dObject";

interface IR2d {
  container: HTMLElement;
  zoom: boolean;
  pan: boolean;
  scale: number;
  ros: IRosBridgeURL | IRos;
}

interface IAddMap {
  image: ImageObject;
  info: {
    naturalHeight: number;
    origin: number[];
    resolution: number;
  };
}

interface IPoseWithTopic {
  object: R2dObject;
  topic: string;
  queueLength?: number;
  throttleRate?: number;
}

interface IPoseWithUpdater {
  object: R2dObject;
}

interface IHeader {
  frame_id: "map";
  seq: number;
  stamp: {
    secs: number;
    nsecs: number;
  };
}

interface Ixyz {
  x: number;
  y: number;
  z: number;
}

export interface IPoseMsg {
  header: IHeader;
  pose: {
    position: Ixyz;
    orientation: Ixyz & {
      w: number;
    };
  };
}

export interface IRosBridgeURL {
  url: string;
}

export type IRos = ROSLib.Ros;

interface IStaticObject {
  object: R2dObject;
  x: number;
  y: number;
  rotationInDegClockwise?: number;
}

interface IPath {
  points: Array<{ x: number; y: number }>;
  color: string;
  width: number;
  tension?: number;
}

export default class R2d {
  private readonly container: HTMLElement;
  private readonly stage: Konva.Stage;
  private readonly zoom: boolean;
  private readonly pan: boolean;
  private readonly scale: number;
  private readonly ros: ROSLib.Ros;
  private readonly mapLayer: Konva.Layer;
  private readonly objectsLayer: Konva.Layer;

  public readonly createMap: (args: IMap) => Promise<IMapReturnSuccess>;
  public readonly createCircle: (args: ICircle) => CircleObject;
  public readonly createRectangle: (args: IRectangle) => RectangleObject;
  public readonly createArrow: (args: IArrow) => ArrowObject;
  public readonly createImageSync: (args: IImageSync) => ImageObject;
  public readonly createImageAsync: (
    args: IImageAsync
  ) => Promise<IImageAsyncReturnSuccess>;
  public readonly createPolygon: (
    args: IPolygon & { points: Array<{ x: number; y: number }> }
  ) => PolygonObject;

  constructor(args: IR2d) {
    this.container = args.container;
    this.zoom = args.zoom;
    this.pan = args.pan;
    this.scale = args.scale;
    this.stage = new Konva.Stage({
      container: this.container,
      ...getContainerDimensions(this.container),
      draggable: this.pan
    });
    this.ros = !isRosBridgeURL(args.ros)
      ? args.ros
      : new ROSLib.Ros({ url: args.ros.url });
    this.mapLayer = new Konva.Layer();
    this.objectsLayer = new Konva.Layer();

    this.stage.add(this.mapLayer);
    this.stage.add(this.objectsLayer);

    if (this.zoom) {
      const zoomControl = new ZoomControl(this.stage);
      zoomControl.handleWheelZoom(this.container, 1.1);
      zoomControl.handleTouchZoom();
    }

    this.createCircle = circle.bind({
      layer: this.objectsLayer,
      stage: this.stage,
      scale: this.scale
    });
    this.createRectangle = rectangle.bind({
      layer: this.objectsLayer,
      stage: this.stage,
      scale: this.scale
    });
    this.createArrow = arrow.bind({
      layer: this.objectsLayer,
      stage: this.stage,
      scale: this.scale
    });
    this.createPolygon = polygon.bind({
      layer: this.objectsLayer,
      stage: this.stage,
      scale: this.scale
    });
    this.createImageSync = imageSync.bind({
      layer: this.objectsLayer,
      stage: this.stage,
      scale: this.scale
    });
    this.createImageAsync = imageAsync.bind({
      layer: this.objectsLayer,
      stage: this.stage,
      scale: this.scale
    });
    this.createMap = drawableMap.bind({
      layer: this.objectsLayer,
      stage: this.stage,
      scale: this.scale
    });
  }

  public changePan = (enable: boolean) => {
    this.stage.draggable(enable);
  };

  public addMap = (mapObject: IAddMap) => {
    const { image, info } = mapObject;
    const { origin, resolution, naturalHeight } = info;
    const nativeObject = image.getNativeObject();

    nativeObject.scaleX(resolution * this.scale);
    nativeObject.scaleY(resolution * this.scale);
    nativeObject.offsetX(-1 * origin[0] * (1 / resolution));
    nativeObject.offsetY(naturalHeight + origin[1] * (1 / resolution));
    nativeObject.position({
      x: this.stage.width() / 2,
      y: this.stage.height() / 2
    });
    nativeObject.rotate(-90);

    this.mapLayer.add(nativeObject);
    this.mapLayer.batchDraw();

    return {
      destroy: () => {
        image.destroy();
        this.mapLayer.batchDraw();
      },
      hide: () => {
        image.hide();
      },
      show: () => {
        image.show();
      }
    };
  };

  public addPoseWithUpdater = (pose: IPoseWithUpdater) => {
    const object = pose.object.getNativeObject();
    object.scaleX(this.scale);
    object.scaleY(this.scale);
    this.objectsLayer.add(object);
    object.hide();

    return {
      updatePose: (poseArg: IPoseMsg["pose"]) => {
        this.moveObjectForPose(object, poseArg);
      },
      destroy: () => {
        object.destroy();
        this.objectsLayer.batchDraw();
      }
    };
  };

  public addPoseWithTopic = (pose: IPoseWithTopic) => {
    const object = pose.object.getNativeObject();
    object.scaleX(this.scale);
    object.scaleY(this.scale);
    this.objectsLayer.add(object);
    object.hide();

    const listenerArg: {
      ros: ROSLib.Ros;
      name: string;
      messageType: string;
      queue_length?: number;
      throttle_rate?: number;
    } = {
      ros: this.ros,
      name: pose.topic,
      messageType: "geometry_msgs/PoseStamped"
    };
    !isNil(pose.queueLength)
      ? (listenerArg.queue_length = pose.queueLength)
      : null;
    !isNil(pose.throttleRate)
      ? (listenerArg.throttle_rate = pose.throttleRate)
      : null;

    const listener = new ROSLib.Topic(listenerArg);

    listener.subscribe((message: IPoseMsg) => {
      this.moveObjectForPose(object, message.pose);
    });

    return {
      destroy: () => {
        object.destroy();
        listener.unsubscribe();
        this.objectsLayer.batchDraw();
      }
    };
  };

  private moveObjectForPose(object: Konva.Shape, pose: IPoseMsg["pose"]) {
    object.show();
    const { x: moveX, y: moveY } = pose.position;
    const { gamma } = quaternionToEulerAngle(pose.orientation);
    animateToPoint(
      object,
      this.stage.width() / 2 - moveY * this.scale,
      this.stage.height() / 2 - moveX * this.scale
    );
    const currentAngle = object.rotation() % 360;
    const desiredAngle =
      ((gamma > 0 ? gamma : 2 * Math.PI + gamma) * 180) / Math.PI;
    object.rotate(-1 * (desiredAngle - Math.abs(currentAngle)));
    this.objectsLayer.batchDraw();
  }

  public addStaticObject = (args: IStaticObject) => {
    const { object, x, y, rotationInDegClockwise } = args;
    const nativeObject = object.getNativeObject();
    nativeObject.scaleX(this.scale);
    nativeObject.scaleY(this.scale);
    nativeObject.x(this.stage.width() / 2 - y * this.scale);
    nativeObject.y(this.stage.height() / 2 - x * this.scale);
    !isNil(rotationInDegClockwise)
      ? nativeObject.rotate(rotationInDegClockwise)
      : null;

    this.objectsLayer.add(nativeObject);
    nativeObject.moveToTop();
    this.objectsLayer.batchDraw();

    return {
      destroy: () => {
        nativeObject.destroy();
        this.objectsLayer.batchDraw();
      },
      hide: () => {
        nativeObject.hide();
      },
      show: () => {
        nativeObject.show();
      }
    };
  };

  public addPath = (args: IPath) => {
    const { points: pointsUnformatted, color, width } = args;
    const points: number[] = [];
    map(pointsUnformatted, p => {
      points.push(this.stage.width() / 2 - p.y * this.scale);
      points.push(this.stage.height() / 2 - p.x * this.scale);
    });

    const line = new Konva.Line({
      points,
      stroke: color,
      strokeWidth: width * this.scale,
      x: points[0],
      y: points[1],
      offset: {
        x: points[0],
        y: points[1]
      },
      tension: !isNil(args.tension) ? args.tension : 0
    });

    this.objectsLayer.add(line);
    this.objectsLayer.batchDraw();

    return {
      destroy: () => {
        line.destroy();
        this.objectsLayer.batchDraw();
      },
      hide: () => {
        line.hide();
      },
      show: () => {
        line.show();
      }
    };
  };
}
